@@ -1,12 +1,12 @@
 ;;; bufferlo.el --- Manage frame/tab-local buffer lists -*- lexical-binding: t -*-
 
-;; Copyright (C) 2021-2023 Free Software Foundation, Inc.
+;; Copyright (C) 2021-2024 Free Software Foundation, Inc.
 
 ;; Author: Florian Rommel <mail@florommel.de>
 ;; Maintainer: Florian Rommel <mail@florommel.de>
 ;; Url: https://github.com/florommel/bufferlo
 ;; Created: 2021-09-15
-;; Version: 0.3
+;; Version: 0.6
 ;; Package-Requires: ((emacs "27.1"))
 ;; Keywords: buffer frame tabs local
 
@@ -106,8 +106,53 @@ This is a list of regular expressions that match buffer names."
   :type '(repeat string))
 
 (defcustom bufferlo-ibuffer-bind-local-buffer-filter t
-  "If non-nil, the local buffer filter is bound to \"/ l\" in ibuffer."
-  :type '(repeat string))
+  "If non-nil, bind the local buffer filter and the orphan filter in ibuffer.
+The local buffer filter is bound to \"/ l\" and the orphan filter to \"/ L\"."
+  :type 'boolean)
+
+(defcustom bufferlo-local-scratch-buffer-name "*local scratch*"
+  "Base name of the local scratch buffer.
+Multiple frame/tabs will use `generate-new-buffer-name' (which
+appends \"<N>\" to the name) in order to get a unique buffer.
+
+Local scratch buffers are optional and not used by default.
+Use the following functions to create and work with them:
+`bufferlo-create-local-scratch-buffer',
+`bufferlo-get-local-scratch-buffer',
+`bufferlo-switch-to-local-scratch-buffer',
+and `bufferlo-toggle-local-scratch-buffer'.
+For example, create a dedicated local scratch buffer for all tabs and frames:
+  (setq \\='tab-bar-new-tab-choice #\\='bufferlo-create-local-scratch-buffer)
+  (add-hook \\='after-make-frame-functions
+             #\\='bufferlo-switch-to-local-scratch-buffer)
+
+You can set this to \"*scratch*\"."
+  :type 'string)
+
+(defcustom bufferlo-local-scratch-buffer-initial-major-mode nil
+  "The initial major mode for local scratch buffers.
+If nil, the local scratch buffers' major mode is set to `initial-major-mode'."
+  :type 'function)
+
+(defcustom bufferlo-anywhere-filter '(switch-to-buffer bufferlo-switch-to-buffer)
+  "The functions that use the local buffer list in `bufferlo-anywhere-mode'.
+If `bufferlo-anywhere-filter-type' is set to `exclude', this is an exclude
+filter (i.e., determines the functions that do not use the local buffer list).
+If `bufferlo-anywhere-filter-type' is set to `include' (or any other value),
+this is an include filter.
+The value can either be a list of functions, or t (for all functions),
+or a custom filter function that takes a function symbol as its argument and
+returns whether the probed function should be filtered (non-nil) or
+not-filtered (nil)."
+  :type '(choice (repeat   :tag "Filter specific functions" function)
+                 (const    :tag "All functions" t)
+                 (function :tag "Custom filter function")))
+
+(defcustom bufferlo-anywhere-filter-type 'exclude
+  "Determines whether `bufferlo-anywhere-filter' is an include or exclude filter.
+Set this to `include' or `exclude'."
+  :type '(radio (const :tag "Include filter" include)
+                (const :tag "Exclude filter" exclude)))
 
 (defvar bufferlo--desktop-advice-active nil)
 (defvar bufferlo--desktop-advice-active-force nil)
@@ -519,6 +564,133 @@ Buffers matching `bufferlo-include-buffer-filters' are not removed."
             (bufferlo-remove buffer)))
       (message "Current buffer is not part of a project"))))
 
+(defun bufferlo-find-buffer (buffer-or-name)
+  "Switch to the frame/tab containing BUFFER-OR-NAME in its local list.
+If multiple frame or tabs contain the buffer, interactively prompt
+for the to-be-selected frame and tab.
+This does not select the buffer -- just the containing frame and tab."
+  (interactive "b")
+  (let* ((buffer (get-buffer buffer-or-name))
+         (flatten (lambda (list)
+                    (apply #'append (append list '()))))
+         (search-tabs (lambda (f)
+                        (let ((i 0))
+                          (mapcar
+                           (lambda (tab)
+                             (setq i (1+ i))
+                             (when (bufferlo-local-buffer-p buffer f (1- i) t)
+                               (list f (frame-parameter f 'name)
+                                     (eq f (selected-frame))
+                                     i (cdr (assq 'name tab)))))
+                           (frame-parameter f 'tabs)))))
+         (search-frames (lambda (f)
+                          (unless (frame-parameter f 'no-accept-focus)
+                            (if (frame-parameter f 'tabs)
+                                ;; has tabs
+                                (funcall search-tabs f)
+                              ;; has no tabs
+                              (when (bufferlo-local-buffer-p buffer f nil t)
+                                (list (list f (frame-parameter f 'name)
+                                            (eq f (selected-frame))
+                                            nil nil)))))))
+         (candidates (seq-filter 'identity
+                                 (funcall flatten
+                                          (mapcar
+                                           (lambda (f)
+                                             (funcall search-frames f))
+                                           (frame-list)))))
+         (candidates (mapcar
+                      (lambda (c)
+                        (let ((sel (if (nth 2 c) " [this]" ""))
+                              (frame-name (nth 1 c))
+                              (frame-obj  (nth 0 c))
+                              (tab-index  (nth 3 c))
+                              (tab-name   (nth 4 c)))
+                          (if tab-index
+                              (cons (format "Frame: %s (%s)%s  Tab %s: %s"
+                                            frame-name frame-obj sel
+                                            tab-index tab-name)
+                                    c)
+                            (cons (format "Frame: %s (%s)%s"
+                                          frame-name frame-obj sel)
+                                  c))))
+                      candidates))
+         (selected (if (cdr candidates)
+                       (completing-read
+                        "Select frame/tab: "
+                        candidates
+                        nil t)
+                     (caar candidates)))
+         (selected (assoc selected candidates)))
+    (if (not selected)
+        (message "Orphan: No frame/tab contains buffer '%s'" (buffer-name buffer))
+      (let ((frame (nth 1 selected))
+            (tab-index (nth 4 selected)))
+        (select-frame-set-input-focus frame)
+        (when tab-index
+          (tab-bar-select-tab tab-index))
+        frame))))
+
+(defun bufferlo-find-buffer-switch (buffer-or-name)
+  "Switch to the frame/tab containig BUFFER-OR-NAME and select the buffer.
+This is like `bufferlo-find-buffer' but additionally selects the buffer."
+  (interactive "b")
+  (when (bufferlo-find-buffer buffer-or-name)
+    (switch-to-buffer buffer-or-name)))
+
+(defun bufferlo-get-local-scratch-buffer ()
+  "Get the local scratch buffer or create it if not already existent and return it."
+  (let ((buffer (seq-find (lambda (b)
+                            (string-match-p
+                             (concat
+                              "^"
+                              (regexp-quote bufferlo-local-scratch-buffer-name)
+                              "\\(<[0-9]*>\\)?$")
+                             (buffer-name b)))
+                          (bufferlo-buffer-list nil nil t))))
+    (unless buffer
+      (setq buffer (get-buffer-create
+                    (generate-new-buffer-name bufferlo-local-scratch-buffer-name)))
+      (with-current-buffer buffer
+        (when (eq major-mode 'fundamental-mode)
+	  (funcall (or bufferlo-local-scratch-buffer-initial-major-mode
+                       initial-major-mode
+                       #'ignore)))))
+    buffer))
+
+(defun bufferlo-create-local-scratch-buffer ()
+  "Create a local scratch buffer and return it."
+  (get-buffer-create (generate-new-buffer-name bufferlo-local-scratch-buffer-name)))
+
+(defun bufferlo-switch-to-scratch-buffer (&optional frame)
+  "Switch to the scratch buffer.
+When FRAME is non-nil, switch to the scratch buffer in the specified frame
+instead of the current one."
+  (interactive)
+  (if frame
+      (with-selected-frame frame
+        (switch-to-buffer "*scratch*"))
+    (switch-to-buffer "*scratch*")))
+
+(defun bufferlo-switch-to-local-scratch-buffer (&optional frame)
+  "Switch to the local scratch buffer.
+When FRAME is non-nil, switch to the local scratch buffer in the specified frame
+instead of the current one."
+  (interactive)
+  (if frame
+      (with-selected-frame frame
+        (switch-to-buffer (bufferlo-get-local-scratch-buffer)))
+    (switch-to-buffer (bufferlo-get-local-scratch-buffer))))
+
+(defun bufferlo-toggle-local-scratch-buffer ()
+  "Switch to the local scratch buffer or bury it if it is already selected.
+Creates a new local scratch buffer if none exists for this frame/tab."
+  (interactive)
+  (let ((buffer (bufferlo-get-local-scratch-buffer)))
+    (if (eq buffer (current-buffer))
+        (bury-buffer)
+      (switch-to-buffer buffer))))
+
 (defun bufferlo-switch-to-buffer (buffer &optional norecord force-same-window)
   "Display the BUFFER in the selected window.
 Completion includes only local buffers.
@@ -565,17 +737,41 @@ If the prefix arument is given, include all buffers."
        (revert-buffer))
      buffer)))
 
+(defun bufferlo-list-orphan-buffers ()
+  "Display a list of orphan buffers."
+  (interactive)
+  (display-buffer
+   (let* ((old-buffer (current-buffer))
+          (name "*Orphan Buffer List*")
+	  (buffer (get-buffer-create name)))
+     (with-current-buffer buffer
+       (Buffer-menu-mode)
+       (setq bufferlo--buffer-menu-this-frame (selected-frame))
+       (setq Buffer-menu-files-only nil)
+       (setq Buffer-menu-buffer-list #'bufferlo--get-orphan-buffers)
+       (setq Buffer-menu-filter-predicate nil)
+       (list-buffers--refresh #'bufferlo--get-orphan-buffers old-buffer)
+       (tabulated-list-print)
+       (revert-buffer))
+     buffer)))
+
 (with-eval-after-load 'ibuf-ext
   (define-ibuffer-filter bufferlo-local-buffers
       "Limit current view to local buffers."
     (:description "local buffers" :reader nil)
-    (bufferlo-local-buffer-p buf)))
+    (bufferlo-local-buffer-p buf))
+  (define-ibuffer-filter bufferlo-orphan-buffers
+      "Limit current view to orphan buffers."
+    (:description "orphan buffers" :reader nil)
+    (not (memq buf (bufferlo--get-captured-buffers)))))
 
 (with-eval-after-load 'ibuffer
   (when bufferlo-ibuffer-bind-local-buffer-filter
     (require 'ibuf-ext)
     (define-key ibuffer--filter-map (kbd "l")
-                #'ibuffer-filter-by-bufferlo-local-buffers)))
+                #'ibuffer-filter-by-bufferlo-local-buffers)
+    (define-key ibuffer--filter-map (kbd "L")
+                #'ibuffer-filter-by-bufferlo-orphan-buffers)))
 
 (defun bufferlo-ibuffer (&optional other-window-p noselect shrink)
   "Invoke `ibuffer' filtered for local buffers.
@@ -592,6 +788,128 @@ The parameters OTHER-WINDOW-P NOSELECT SHRINK are passed to `ibuffer'."
     (ibuffer other-window-p name '((bufferlo-local-buffers . nil))
              noselect shrink)))
 
+(defun bufferlo-ibuffer-orphans (&optional other-window-p noselect shrink)
+  "Invoke `ibuffer' filtered for orphan buffers.
+The parameters OTHER-WINDOW-P NOSELECT SHRINK are passed to `ibuffer'."
+  (interactive)
+  (let ((name "*Bufferlo Orphans Ibuffer*"))
+    (ibuffer other-window-p name '((bufferlo-orphan-buffers . nil))
+             noselect shrink)))
+
+(define-minor-mode bufferlo-anywhere-mode
+  "Frame/tab-local buffer lists anywhere you like.
+Enables bufferlo's local buffer list for any function that interactively prompts
+for buffers via `read-buffer'.  By default this enables the local buffer list
+for (almost) all functions.  Customize `bufferlo-anywhere-filter' and
+`bufferlo-anywhere-filter-type' to adapt the behavior.
+This minor mode requires `bufferlo-mode' to be enabled.
+You can use `bufferlo-anywhere-disable' to disable the local buffer list for
+the next command, when the mode is enabled."
+  :global t
+  :require 'bufferlo
+  :init-value nil
+  :lighter nil
+  :keymap nil
+  (if bufferlo-anywhere-mode
+      (advice-add #'call-interactively :around #'bufferlo--interactive-advice)
+    (advice-remove #'call-interactively #'bufferlo--interactive-advice)))
+
+(defvar bufferlo--anywhere-tmp-enabled nil)
+(defvar bufferlo--anywhere-tmp-disabled nil)
+(defvar bufferlo--anywhere-old-read-buffer-function nil)
+(defvar bufferlo--anywhere-nested nil)
+
+(defun bufferlo--interactive-advice (oldfn function &optional record-flags keys)
+  "Advice function for `call-interactively' for `bufferlo-anywhere-mode'.
+Temporarily overrides the `read-buffer-function' to filter the
+available buffers to bufferlo's local buffer list.
+OLDFN is the original function.
+FUNCTION is the interactively called functions.
+RECORD-FLAGS and KEYS are passed to `call-interactively'."
+  (if (or bufferlo--anywhere-tmp-enabled
+          (and (not bufferlo--anywhere-tmp-disabled)
+               (xor (eq bufferlo-anywhere-filter-type 'exclude)
+                    (cond
+                     ((eq bufferlo-anywhere-filter t) t)
+                     ((listp bufferlo-anywhere-filter)
+                      (memq function bufferlo-anywhere-filter))
+                     ((functionp bufferlo-anywhere-filter)
+                      (funcall bufferlo-anywhere-filter function))))))
+      (let* ((bufferlo--anywhere-old-read-buffer-function
+              ;; Preserve the original `read-buffer-function' but not for
+              ;; nested calls; otherwise we would save our own function.
+              (if bufferlo--anywhere-nested
+                  bufferlo--anywhere-old-read-buffer-function
+                read-buffer-function))
+             (bufferlo--anywhere-nested t)
+             (read-buffer-function
+              (lambda (prompt &optional def require-match predicate)
+                (let ((read-buffer-function
+                       bufferlo--anywhere-old-read-buffer-function))
+                  (read-buffer prompt def require-match
+                               (lambda (b)
+                                 (and (bufferlo-local-buffer-p
+                                       (get-buffer
+                                        (if (stringp b) b (car b))))
+                                      (or (not predicate)
+                                          (funcall predicate b)))))))))
+        (apply oldfn (list function record-flags keys)))
+    ;; `call-interactively' can be nested, e.g., on M-x invocations.
+    ;; Therefore, we restore the original value of the `read-buffer-function'
+    ;; if we do not use bufferlo's local buffer list for this call.
+    (let ((read-buffer-function
+           (if bufferlo--anywhere-nested
+               bufferlo--anywhere-old-read-buffer-function
+             read-buffer-function)))
+      (apply oldfn (list function record-flags keys)))))
+
+(defun bufferlo-anywhere-disable-prefix ()
+  "Disable `bufferlo-anywhere-mode' only for the next command.
+Has no effect if `bufferlo-anywhere-mode' is not enabled.
+Has no effect if the next command does not query for a buffer."
+  (interactive)
+  (let* ((command this-command)
+         (minibuffer-depth (minibuffer-depth))
+         (postfun (make-symbol "bufferlo--anywhere-reenable-next-command")))
+    (fset postfun
+          (lambda ()
+            (unless (or
+                     ;; from window.el:display-buffer-override-next-command
+		     (> (minibuffer-depth) minibuffer-depth)
+		     (eq this-command command))
+              (setq bufferlo--anywhere-tmp-disabled nil)
+              (remove-hook 'post-command-hook postfun))))
+    (setq bufferlo--anywhere-tmp-disabled t)
+    (add-hook 'post-command-hook postfun)))
+
+(defun bufferlo-anywhere-enable-prefix ()
+  "Use bufferlo's local buffer list for the next command.
+Has a similar effect as `bufferlo-anywhere-mode' but only for the next command.
+Has no effect if the next command does not query for a buffer.
+Can be used with or without `bufferlo-anywhere-mode' enabled.
+In contrast to `bufferlo-anywhere-mode', this does not adhere to
+`bufferlo-anywhere-filter'.  Therefore, you can use it in conjunction with
+`bufferlo-anywhere-mode' to temporarily disable the configured filters."
+  (interactive)
+  (let* ((command this-command)
+         (minibuffer-depth (minibuffer-depth))
+         (postfun (make-symbol "bufferlo--anywhere-disable-next-command")))
+    (fset postfun
+          (lambda ()
+            (unless (or
+                     ;; from window.el:display-buffer-override-next-command
+		     (> (minibuffer-depth) minibuffer-depth)
+		     (eq this-command command))
+              (setq bufferlo--anywhere-tmp-enabled nil)
+              (unless bufferlo-anywhere-mode
+                (advice-remove #'call-interactively
+                               #'bufferlo--interactive-advice))
+              (remove-hook 'post-command-hook postfun))))
+    (setq bufferlo--anywhere-tmp-enabled t)
+    (unless bufferlo-anywhere-mode
+      (advice-add #'call-interactively :around #'bufferlo--interactive-advice))
+    (add-hook 'post-command-hook postfun)))
+
 (provide 'bufferlo)
 
 ;;; bufferlo.el ends here
